package tab;

import addr.addrIP;
import java.util.ArrayList;
import java.util.List;

/**
 * represents one route policy handler
 *
 * @author matecsaba
 */
public class tabRtrplc extends tabListing<tabRtrplcN, addrIP> {

    /**
     * reindent table
     */
    public void reindent() {
        int o = 0;
        for (int i = 0; i < entries.size(); i++) {
            tabRtrplcN ntry = entries.get(i);
            if (ntry == null) {
                continue;
            }
            ntry.indent = o;
            switch (ntry.doMode) {
                case iff:
                    o++;
                    break;
                case elsif:
                case els:
                    ntry.indent--;
                    break;
                case enif:
                    ntry.indent--;
                    o--;
                    break;
            }
        }
    }

    private int findEnif(int pos, int lvl) {
        pos++;
        for (; pos < entries.size(); pos++) {
            tabRtrplcN ntry = entries.get(pos);
            if (ntry == null) {
                continue;
            }
            if (ntry.indent > lvl) {
                continue;
            }
            if (ntry.doMode == tabRtrplcN.doType.enif) {
                break;
            }
        }
        return pos;
    }

    /**
     * update one entry
     *
     * @param net network number
     * @param copy copy before update
     * @return null if denied, copy otherwise
     */
    public tabRouteEntry<addrIP> dorpl(tabRouteEntry<addrIP> net, boolean copy) {
        if (copy) {
            net = net.copyBytes();
        }
        int lvl = 0;
        for (int pos = 0; pos < entries.size(); pos++) {
            tabRtrplcN ntry = entries.get(pos);
            if (ntry == null) {
                continue;
            }
            if (ntry.indent > lvl) {
                continue;
            }
            switch (ntry.doMode) {
                case next:
                    continue;
                case description:
                    continue;
                case iff:
                    if (ntry.matches(net)) {
                        lvl++;
                        continue;
                    }
                    pos++;
                    for (; pos < entries.size(); pos++) {
                        for (; pos < entries.size(); pos++) {
                            ntry = entries.get(pos);
                            if (ntry == null) {
                                continue;
                            }
                            if (ntry.indent > lvl) {
                                continue;
                            }
                            break;
                        }
                        ntry = entries.get(pos);
                        if (ntry == null) {
                            break;
                        }
                        boolean stop = false;
                        switch (ntry.doMode) {
                            case elsif:
                                if (ntry.matches(net)) {
                                    lvl++;
                                    stop = true;
                                    break;
                                }
                                break;
                            case els:
                                lvl++;
                                stop = true;
                                break;
                            default:
                                stop = true;
                                break;
                        }
                        if (stop) {
                            break;
                        }
                    }
                    continue;
                case elsif:
                    lvl--;
                    pos = findEnif(pos, lvl);
                    continue;
                case els:
                    lvl--;
                    pos = findEnif(pos, lvl);
                    continue;
                case enif:
                    lvl--;
                    continue;
                case pass:
                    return net;
                case drop:
                    return null;
                case tcl:
                    List<String> scr = new ArrayList<String>();
                    for (; pos < entries.size(); pos++) {
                        ntry = entries.get(pos);
                        if (ntry == null) {
                            continue;
                        }
                        if (ntry.doMode != tabRtrplcN.doType.tcl) {
                            break;
                        }
                        scr.add(ntry.strVal);
                    }
                    pos--;
                    tabRtrmapN.doTcl(net, scr);
                    continue;
                default:
                    ntry.update(net);
                    continue;
            }
        }
        return null;
    }

}
